{ Unit: BcExplorerPanel
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2000-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2000-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    TBcExplorerPanel component. Look and feel of Explorer Band in Microsoft's
    Internet Explorer and Outlook.

    Component works like normal container. Like normal panel, difference
    is that it has paneled caption and close button.

    Properties:
      Caption: string        -- Caption for panel (normally you set this to none)
      BandCaption: string    -- Top Caption
      AutoHide: Boolean      -- If this is true panel is hidden when Close button
                                is pressed. Can be overridden in
                                OnCloseClick event
      BandStyle: TBandStyles -- Change caption panel place
      BandPopupMenu: TPopupMenu -- Caption panel's popupmenu
      Splitter               -- You can assign splitter which is kind of linked.
                                So it gets hidden if explore panel is hidden too
      CloseButtonVisible     -- Is the Close X button visible?

    Methods:
      Show, Hide             -- Changes visible value and if splitter is assigned
                                it's visible status changes too. (you may use
                                Visible property as well)
    Events:
      OnCloseClick           -- called when Close button clicked. If AllowClose
                                is set true then panel is hidden after event.
                                AllowClose defaults to AutoHide.


  History:

    2001-10-12: Visible property supports the Splitter link now. It fixes
                the CloseButton click autoclose and splitter link issue too
    2001-10-11: Changed to TBcExplorerPanel and added to Bluecave Components
                Fixed few issues on BandCaption
                Added CloseButtonVisible property. Note: Change can be
                seen only in Runtime.
    2000-05-29: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcExplorerPanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Menus,
  ExtCtrls, Buttons;

const
  CloseButtonWidth = 20;
  CloseButtonHeight = 17;
  BandCaptionHeight = 22;
  DefaultCloseButtonVisible = True;

type
  TBandStyles = (psVerticalTop, psVerticalBottom, psHorizontalLeft, psHorizontalRight);

  TCloseButtonEvent = procedure(Sender: TObject; var AllowClose: Boolean) of object;

  TBcExplorerPanel = class(TPanel)
  private
    { Private declarations }
    FCaptionPanel: TPanel;
    FCloseButton: TSpeedButton;

    FAutoHide: Boolean;

    FOnCloseClick: TCloseButtonEvent;

    FCaption: string;
    FBandStyle: TBandStyles;
    FSplitter: TSplitter;

    procedure SetCaption(Value: string);
    procedure SetBandStyle(Value: TBandStyles);

    procedure SetBandPopupMenu(Value: TPopupMenu);
    function GetBandPopupMenu: TPopupMenu;

    procedure UpdateStyle;

    procedure OnCloseButtonPress(Sender: TObject);
    function GetCloseButtonVisible: Boolean;
    procedure SetCloseButtonVisible(const Value: Boolean);
    function GetVisible: Boolean;
    procedure SetVisible(const Value: Boolean);
  protected
    { Protected declarations }
    procedure Loaded; override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Hide;
    procedure Show;
  published
    { Published declarations }
    property BandCaption: string read FCaption write SetCaption;
    property Visible: Boolean read GetVisible write SetVisible;
    property AutoHide: Boolean read FAutoHide write FAutoHide default False;
    property BandStyle: TBandStyles read FBandStyle write SetBandStyle default psVerticalTop;
    property Splitter: TSplitter read FSplitter write FSplitter;
    property BandPopupMenu: TPopupMenu read GetBandPopupMenu write SetBandPopupMenu;
    property CloseButtonVisible: Boolean read GetCloseButtonVisible write SetCloseButtonVisible;
    property OnCloseClick: TCloseButtonEvent read FOnCloseClick write FOnCloseClick;
  end;

implementation

constructor TBcExplorerPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FBandStyle := psVerticalTop;

  Height := 280; { default height }

  BevelOuter := bvLowered;
  FCaptionPanel := TPanel.Create(Self);
  with FCaptionPanel do
  begin
    BevelOuter := bvRaised;
    Alignment := taLeftJustify;

    Parent := Self;
    ParentColor := True;
    ParentShowHint := True;
    ParentFont := True;

    Align := alTop;
    Height := BandCaptionHeight;
  end;

  FCloseButton := TSpeedButton.Create(FCaptionPanel);
  with FCloseButton do
  begin
    Anchors := [akTop, akRight];
    SetBounds(Self.Width - CloseButtonWidth - 3, 2, CloseButtonWidth, CloseButtonHeight);

    Margin := 2;
    Layout := blGlyphBottom;

    Flat := True;
    Transparent := False;

    Font.Name := 'Marlett';
    Font.Size := 7;
    Caption := 'r'; { X }
    OnClick := OnCloseButtonPress;

    Visible := DefaultCloseButtonVisible;

    Parent := FCaptionPanel;
  end;

//  Update;

  FAutoHide := False;
  FCaption := '';
end;

destructor TBcExplorerPanel.Destroy;
begin
  FCloseButton.Free;
  FCaptionPanel.Free;

  inherited Destroy;
end;

procedure TBcExplorerPanel.SetCaption(Value: string);
begin
  if (FCaption <> Value) then
  begin
    FCaption := Value;
    FCaptionPanel.Caption := '  ' + Value;
  end;
end;

procedure TBcExplorerPanel.SetBandStyle(Value: TBandStyles);
begin
  if FBandStyle <> Value then
  begin
    FBandStyle := Value;
    UpdateStyle;
  end;
end;
procedure TBcExplorerPanel.UpdateStyle;
begin
  case FBandStyle of
    psVerticalTop:
      begin
        with FCaptionPanel do
        begin
          Alignment := taLeftJustify;
          Align := alTop;
          Height := BandCaptionHeight;
        end;

        with FCloseButton do
        begin
          Anchors := [akTop, akRight];
          SetBounds(Self.Width - CloseButtonWidth - 3,2,CloseButtonWidth,CloseButtonHeight);
        end;
      end;
    psVerticalBottom:
      begin
        with FCaptionPanel do
        begin
          Alignment := taLeftJustify;
          Align := alBottom;
          Height := BandCaptionHeight;
        end;

        with FCloseButton do
        begin
          Anchors := [akTop, akRight];
          SetBounds(Self.Width - CloseButtonWidth - 3,2,CloseButtonWidth,CloseButtonHeight);
        end;
      end;
    psHorizontalLeft:
      begin
        with FCaptionPanel do
        begin
          Alignment := taLeftJustify;
          Align := alLeft;
          Width := BandCaptionHeight;
        end;

        with FCloseButton do
        begin
          Anchors := [akTop, akRight];
          SetBounds(1,2,CloseButtonWidth,CloseButtonHeight);
        end;
      end;
    psHorizontalRight:
      begin
        with FCaptionPanel do
        begin
          Alignment := taLeftJustify;
          Align := alRight;
          Width := BandCaptionHeight;
        end;

        with FCloseButton do
        begin
          Anchors := [akTop, akRight];
          SetBounds(1,2,CloseButtonWidth,CloseButtonHeight);
        end;
      end;
  end;
end;

procedure TBcExplorerPanel.SetBandPopupMenu(Value: TPopupMenu);
begin
  FCaptionPanel.PopupMenu := Value;
end;

function TBcExplorerPanel.GetBandPopupMenu: TPopupMenu;
begin
  Result := FCaptionPanel.PopupMenu;
end;

procedure TBcExplorerPanel.Hide;
begin
  Visible := False;
end;

procedure TBcExplorerPanel.Show;
begin
  Visible := True;
end;

procedure TBcExplorerPanel.OnCloseButtonPress(Sender: TObject);
var AllowClose: Boolean;
begin
  AllowClose := FAutoHide;

  if Assigned(FOnCloseClick) then FOnCloseClick(Self, AllowClose);

  if AllowClose then
  begin
    Self.Visible := False;
  end;
end;

function TBcExplorerPanel.GetCloseButtonVisible: Boolean;
begin
  Result := FCloseButton.Visible;
end;

procedure TBcExplorerPanel.SetCloseButtonVisible(const Value: Boolean);
begin
  FCloseButton.Visible := Value;
end;

procedure TBcExplorerPanel.Loaded;
begin
  inherited Loaded;
end;

function TBcExplorerPanel.GetVisible: Boolean;
begin
  Result := inherited Visible;
end;

procedure TBcExplorerPanel.SetVisible(const Value: Boolean);
begin
  if Assigned(FSplitter) then
    FSplitter.Visible := Value;
  inherited Visible := Value;
end;

end.
