{ Unit: BcBlockParser
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Parses semi-script blocks with OR operator and replace tags.

    Example with default block/tag chars (quota chars not in the example):

      "Playing track: [<artist>=|No artist|<artist>] - [<title>=|No title|<title>]"


    If <artist>" = "Jarre" and <title> = "Oxygene 4" it will parse to:

      "Playing track: Jarre - Oxygene 4"


    But if <title> = "" it will parse to:

      "Playing track: Jarre - No title"

    There is simple boolean True/False comparsion too:

    "[<someboolean>|It was True|It was False]"

    In this example if <someboolean> contains text "true" it will be true, otherwise
    false. Note the missing "=" for this boolean comparsion.

    There are two fixed internal tags, <TRUE> and <FALSE>. Their values are "TRUE"
    and "FALSE". Thus the comparsion above could be "[<someboolean>=<TRUE>|...".
    Actually if the "=<TRUE>" is omitted the compiler will assume that string as
    comparsion.

    As the example shows the first part in the block OR operator (defaults to
    pipe char) is the Expression which will result True if Expression is
    not empty and False if Expression is empty.

      [Expression|True|False]

    To use one of the block chars as a character you must escape it with '\'.
    To have a backslash character put double backslash, '\\'. To add new line
    use '\n'.

    You can also nest the blocks. Consider this:

      "\<someboolean\> = [<someboolean>|[<useyep>|Yep|True]|False]\n"


    If <someboolean> = "true" and <useyep> = "true" then it will parse to:

      "<someboolean> = Yep", with new line (#13#10) at the end.


    More examples:

      "\<number\> = [<number>=0|It is Zero|<number>]\n"
      "[<somestr>=||Not empty and it is '<somestr>']"


    Note: By default comparsions are not case-sensitive. Tag names should
      not be case sensitive at all. Keep this in mind when providing tag
      values through the event. All tags are internally stored in upper case.


  History:

    2001-11-13: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcExpressions;

interface

uses Classes, SysUtils;

const
  DefaultTagBeginChar = '<';
  DefaultTagEndChar = '>';
  DefaultBlockBeginChar = '[';
  DefaultBlockEndChar = ']';
  DefaultBlockORChar = '|';
  DefaultBlockCmpChar = '=';
  DefaultEscapeChar = '\';
  DefaultEnabled = False;
  DefaultCaseSensitive = False;
  DefaultRaiseErrors = True;

const
  OPNil = 0;
  OPHalt = 1;
  OPExit = 2;
  OPConcat = 3;
  OPJmp = 4;
  OPCmp = 5;
  OPClr = 6;

type
  EExpressionCompileError = class(Exception);

  TBcExpressionCodeSegment = class;

  TOnTagEvent = procedure(ATag: string; var AValue: string) of object;
  TOnTagUpdateEvent = function(ATag: string; var AValue: string): Boolean of object;

  TBcExpressionCompiler = class(TObject)
  private
    FInput: string;
    FEnabled: Boolean;
    FBlockORChar: Char;
    FBlockBeginChar: Char;
    FBlockEndChar: Char;
    FTagEndChar: Char;
    FTagBeginChar: Char;
    FCodeSegment: TBcExpressionCodeSegment;
    FBlockCmpChar: Char;
    FEscapeChar: Char;
    FRaiseErrors: Boolean;
    procedure SetInput(const Value: string);
    procedure SetEnabled(const Value: Boolean);

    function FindNextChar(ALine: string; AChar: Char; P: Integer; HandleEscape: Boolean): Integer; virtual;
    function FindNextPairedChar(ALine: string; AChar1, AChar2: Char; P: Integer): Integer; virtual;
    procedure ProcessOR(ALine: string; var Exp1, Exp2, Result1, Result2: string);
    function ParseEscapedChars(ALine: string): string;
    function GetCompilePoint: Integer;
    function GetCaseSensitive: Boolean;
    procedure SetCaseSensitive(const Value: Boolean);

  public
    constructor Create; virtual;
    destructor Destroy; override;

    function Compile: Boolean;

    procedure ProcessLine(ALine: string);
    procedure ProcessTags(ALine: string; Target: Integer; AClear: Boolean);

    property CompilePoint: Integer read GetCompilePoint;
    property CodeSegment: TBcExpressionCodeSegment read FCodeSegment{ write FCodeSegment};
    property Input: string read FInput write SetInput;
    property BlockBeginChar: Char read FBlockBeginChar write FBlockBeginChar default DefaultBlockBeginChar;
    property BlockEndChar: Char read FBlockEndChar write FBlockEndChar default DefaultBlockEndChar;
    property BlockORChar: Char read FBlockORChar write FBlockORChar default DefaultBlockORChar;
    property BlockCmpChar: Char read FBlockCmpChar write FBlockCmpChar default DefaultBlockCmpChar;
    property EscapeChar: Char read FEscapeChar write FEscapeChar default DefaultEscapeChar;
    property TagBeginChar: Char read FTagBeginChar write FTagBeginChar default DefaultTagBeginChar;
    property TagEndChar: Char read FTagEndChar write FTagEndChar default DefaultTagEndChar;
    property Enabled: Boolean read FEnabled write SetEnabled;
    property RaiseErrors: Boolean read FRaiseErrors write FRaiseErrors default DefaultRaiseErrors;
    property CaseSensitive: Boolean read GetCaseSensitive write SetCaseSensitive default DefaultCaseSensitive;
  end;

  PBcCodeBlock = ^TBcCodeBlock;
  TBcCodeBlock = record
    OPCode: Integer;
    case {OPCodeParameters:} Integer of // max params size is 6 integers
      OPNil: (Param0, Param1, Param2, Param3: Integer;
         Param4, Param5, Param6, Param7: SmallInt);          { 4+4+4+4+2+2+2+2 }
      OPHalt: (ExitCode: Integer);                           { 4 }
      OPExit: (Result: Integer);                             { 4 }
      OPConcat: (Source1, Source2, Target: Integer);         { 4+4+4 }
      OPJmp: (JmpOffset: Integer);                           { 4 }
      OPCmp: (Exp1, Exp2, JumpEqual, JumpNotEqual: Integer); { 4+4+4+4 }
      OPClr: (Index: Integer);
  end;

  TBcCodeBlocks = array of TBcCodeBlock;

  PBcDataString = ^TBcDataString;
  TBcDataString = record
    Reference: Integer; // -2 no reference, -1 reference, but not updated
    Value: string;
  end;

  TBcDataStrings = array of TBcDataString;

  TBcExpressionDataSegment = class(TObject)
  private
//    FLineBreak: Integer;
    FVar1: Integer;
    FVar2: Integer;
    FResultLine: Integer;

    FDataStrings: TBcDataStrings;
    FRuntimeStrings: TBcDataStrings;
    FOnTag: TOnTagUpdateEvent;
    function DoOnTag(ATag: string; var AValue: string): Boolean;
    function GetCount: Integer;
    function GetStrings(Index: Integer): PBcDataString;
    procedure AddPredefinedRuntimeData;
    function GetResultValue: string;
    function GetRuntimeCount: Integer;
  public
    constructor Create; virtual;
    destructor Destroy; override;
//    procedure LoadFromStream(Source: TStream);
//    procedure SaveToStream(Target: TStream);

    procedure Clear;
    procedure ClearResultValue;
    function Add(Value: string; Reference: Integer): Integer;
    function AddRuntime(Value: string; Reference: Integer): Integer;
    function IndexOfTag(Value: string; var AResult: Integer): Boolean;
    function GetTagValue(Index: Integer): string;
    property Count: Integer read GetCount;
    property RuntimeCount: Integer read GetRuntimeCount;
    property ResultValue: string read GetResultValue;
    property Strings[Index: Integer]: PBcDataString read GetStrings; default;
    property OnTag: TOnTagUpdateEvent read FOnTag write FOnTag;
  end;

  TBcExpressionCodeSegment = class(TObject)
  private
    FDataSegment: TBcExpressionDataSegment;
    FCodeBlocks: TBcCodeBlocks;
    FRunPoint: Integer;
    FCaseSensitive: Boolean;
    function GetCount: Integer;
    function GetOutput: string;
    function GetCodeBlocks(Index: Integer): PBcCodeBlock;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function Add({var} CodeBlock: TBcCodeBlock): Integer;
    function AddJmp(JmpOffset: Integer): Integer;
    function AddClr(Index: Integer): Integer;
    function AddConcat(Source1, Source2, Target: Integer): Integer;
//    procedure LoadFromStream(Source: TStream);
//    procedure SaveToStream(Target: TStream);
    procedure Clear;
    function Execute: Integer;
    function ExecuteNext: Integer;
    property Output: string read GetOutput;
    property RunPoint: Integer read FRunPoint write FRunPoint;
    property Count: Integer read GetCount;
    property CodeBlocks[Index: Integer]: PBcCodeBlock read GetCodeBlocks; default;
    property DataSegment: TBcExpressionDataSegment read FDataSegment;
    property CaseSensitive: Boolean read FCaseSensitive write FCaseSensitive default DefaultCaseSensitive;
  end;

implementation

uses Dialogs;

constructor TBcExpressionCompiler.Create;
begin
  FTagBeginChar :=DefaultTagBeginChar;
  FTagEndChar := DefaultTagEndChar;
  FBlockBeginChar := DefaultBlockBeginChar;
  FBlockEndChar := DefaultBlockEndChar;
  FBlockORChar := DefaultBlockORChar;
  FBlockCmpChar := DefaultBlockCmpChar;
  FEscapeChar := DefaultEscapeChar;
  FEnabled := DefaultEnabled;
  FRaiseErrors := DefaultRaiseErrors;
  FCodeSegment := TBcExpressionCodeSegment.Create;
end;

destructor TBcExpressionCompiler.Destroy;
begin
  FreeAndNil(FCodeSegment);
  inherited Destroy;
end;

function TBcExpressionCompiler.FindNextChar(ALine: string; AChar: Char; P: Integer;  HandleEscape: Boolean): Integer;
var
  I: Integer;
  Escaped: Boolean;
  C: Char;
begin
  Result := 0;
  if Length(ALine) = 0 then
    Exit;
  I := P;
  Escaped := False;
  while I <= Length(ALine) do
  begin
    if not Escaped then
    begin
      C := ALine[I];
      if HandleEscape and (C = FEscapeChar) then
        Escaped := True
      else
      begin
        if C = AChar then
        begin
          Result := I;
          Break;
        end;
      end;
    end else
      Escaped := False;
    Inc(I, 1);
  end;
end;

function TBcExpressionCompiler.FindNextPairedChar(ALine: string; AChar1, AChar2: Char; P: Integer): Integer;
var
  I: Integer;
  Escaped, StartFound: Boolean;
  C: Char;
begin
  Result := 0;
  if Length(ALine) = 0 then
    Exit;
  I := P;
  Escaped := False;
  StartFound := False;
  while I <= Length(ALine) do
  begin
    if not Escaped then
    begin
      C := ALine[I];
      if C = FEscapeChar then
        Escaped := True
      else
      begin
        if C = AChar1 then
        begin
          Inc(Result, 1);
          StartFound := True;
        end else if C = AChar2 then
          Dec(Result, 1);
        if StartFound and (Result = 0) then
        begin
          Result := I;
          //Break;
          Exit;
        end;
      end;
    end else
      Escaped := False;
    Inc(I, 1);
  end;
  if Result > 0 then
  begin
//    Result := P;
    if FRaiseErrors then
      raise EExpressionCompileError.Create('Expecting block end but end of expression found.');
  end;
  Result := P;
end;

procedure TBcExpressionCompiler.ProcessLine(ALine: string);
var
  I, B, E, CI: Integer;
  S, Exp1, Exp2, Result1, Result2: string;
  C: TBcCodeBlock;
begin
  if Length(ALine) = 0 then
    Exit;
  I := 1;
  E := 0;
  while True do
  begin
    B := FindNextChar(ALine, FBlockBeginChar, I, True);
    if B = 0 then
      Break;
    S := Copy(ALine, I, B - I); // copy chars before begin
    ProcessTags(S, FCodeSegment.FDataSegment.FResultLine, False); // process tags/strings to execution concat codes

    E := FindNextPairedChar(ALine, FBlockBeginChar, FBlockEndChar, B); // find end
    S := Copy(ALine, B + 1, E - B - 1);
    ProcessOR(S, Exp1, Exp2, Result1, Result2);     // parse OR block
    ProcessTags(Exp1, FCodeSegment.FDataSegment.FVar1, True); // generate code for expression
    ProcessTags(Exp2, FCodeSegment.FDataSegment.FVar2, True);
    FillChar(C, SizeOf(C), 0);                      // add cmp opcode
    C.OPCode := OPCmp;
    C.Exp1 := FCodeSegment.FDataSegment.FVar1;
    C.Exp2 := FCodeSegment.FDataSegment.FVar2;
    C.JumpEqual := 0;                               // no jump
    C.JumpNotEqual := 0;                            // updated after equal expression length has been parsed
    CI := FCodeSegment.Add(C);

    ProcessLine(Result1);                           // parse equal string
    FCodeSegment.FCodeBlocks[CI].JumpNotEqual := CompilePoint - CI; // not equal jmp over equal (+1 opjmp for equal over not equal)
    CI := FCodeSegment.AddJmp(0);
    ProcessLine(Result2);
    FCodeSegment.FCodeBlocks[CI].JmpOffset := CompilePoint - CI - 1; // equal end jmp over not equal
    I := E + 1;
  end;
  ProcessTags(Copy(ALine, E + 1, Length(ALine)), FCodeSegment.FDataSegment.FResultLine, False); // end strings
end;

procedure TBcExpressionCompiler.ProcessOR(ALine: string; var Exp1, Exp2, Result1,
 Result2: string);
var
  I, P: Integer;
  Expression: string;
begin
  Exp1 := '';
  Exp2 := '';
  Result1 := '';
  Result2 := '';
  if Length(ALine) = 0 then
    Exit;

  // Length 0 checked above
  I := FindNextChar(ALine, FBlockORChar, 1, True);
  if I = 0 then
    Exit;
  Expression := Copy(ALine, Length(FBlockBeginChar), I - Length(FBlockBeginChar));

  // first
  P := I; // skip blocks
  I := FindNextChar(ALine, FBlockORChar,
    FindNextPairedChar(ALine, FBlockBeginChar, FBlockEndChar, I) + 1, True);
  if I = 0 then
    I := Length(ALine) + 1;
  Result1 := Copy(ALine, P + 1, I - P - 1);

  // latter
//  I := FindNextChar(ALine, FBlockORChar, I);
  if I > 0 then
    Result2 := Copy(ALine, I + 1, Length(ALine));

  // separate expressions
  I := FindNextChar(Expression, FBlockCmpChar, 1, True);
  if I = 0 then // no comparing, compare to empty string
  begin
    Exp1 := Copy(Expression, 1, Length(Expression));
    Exp2 := FTagBeginChar + 'TRUE' + FTagEndChar; // fixed comparsion to TRUE tag
  end else
  begin
    Exp1 := Copy(Expression, 1, I - 1);
    Exp2 := Copy(Expression, I + 1, Length(Expression));
  end;
end;

procedure TBcExpressionCompiler.SetEnabled(const Value: Boolean);
begin
  FEnabled := Value;
  if FEnabled then
    Compile;
end;

procedure TBcExpressionCompiler.SetInput(const Value: string);
begin
  FInput := Value;
  if FEnabled then
    Compile;
end;

procedure TBcExpressionCompiler.ProcessTags(ALine: string; Target: Integer; AClear: Boolean);
var
  B, E, I: Integer;
  S: string;
  C: TBcCodeBlock;
begin
  I := 1;
  E := 0;
  if AClear then
  begin
    FillChar(C, SizeOf(C), 0); // add clr opcode
    C.OPCode := OPClr;
    C.Index := Target;
    FCodeSegment.Add(C);
  end;
  while True do
  begin
    B := FindNextChar(ALine, FTagBeginChar, I, True); // returns 0 if not found or line empty
    if B = 0 then
      Break;

    S := Copy(ALine, I, B - I); // copy chars before begin
    S := ParseEscapedChars(S);
    if S <> '' then
      FCodeSegment.AddConcat(Target, FCodeSegment.FDataSegment.Add(S, -2), Target);

    I := FindNextChar(ALine, FTagBeginChar, B + 1, True);
    E := FindNextChar(ALine, FTagEndChar, B + 1, True);
    if (E = 0) or ((I > 0) and (I < E)) then // next tag begins before new
    begin
      if FRaiseErrors then
      begin
        I := FindNextChar(ALine, FEscapeChar, B + 1, False);
{        if I = 0 then
          I := FindNextChar(ALine, FBlockBeginChar, B + 1);
        if I = 0 then
          I := FindNextChar(ALine, FBlockEndChar, B + 1);}
        if I = 0 then
          I := Length(ALine)
        else
          I := I - B;
        raise EExpressionCompileError.CreateFmt('Unmatched tag found ''%s''.', [Copy(ALine, B, I)]);
      end;
      Break;
    end;
    S := Copy(ALine, B + 1, E - B - 1); // get tag name
    if S <> '' then
      FCodeSegment.AddConcat(Target, FCodeSegment.FDataSegment.Add(UpperCase(S), -1), Target);

    I := E + 1;
  end;
  S := Copy(ALine, E + 1, Length(ALine)); // copy chars after last tag end
  S := ParseEscapedChars(S);
  if S <> '' then
    FCodeSegment.AddConcat(Target, FCodeSegment.FDataSegment.Add(S, -2), Target);
end;

{ TBcExpressionDataSegment }

function TBcExpressionDataSegment.Add(Value: string; Reference: Integer): Integer;
begin
//  Result := -1;
  try
    if Reference = -1 then  // try to find existing tag
    begin
      if IndexOfTag(Value, Result) then
        Exit;
    end;

    Result := Length(FDataStrings);
    SetLength(FDataStrings, Result + 1);
    FDataStrings[Result].Reference := Reference;
    FDataStrings[Result].Value := Value;
  finally
  end;
end;

function TBcExpressionDataSegment.AddRuntime(Value: string;
  Reference: Integer): Integer;
begin
//  Result := -1;
  try
    Result := Length(FRuntimeStrings);
    SetLength(FRuntimeStrings, Result + 1);
    FRuntimeStrings[Result].Value := Value;
    FRuntimeStrings[Result].Reference := Reference;
    Result := -Result;
  finally
  end;
end;

procedure TBcExpressionDataSegment.Clear;
begin
  FDataStrings := nil;
  FRuntimeStrings := nil;
  AddPredefinedRuntimeData;
end;

constructor TBcExpressionDataSegment.Create;
begin
  //
end;

destructor TBcExpressionDataSegment.Destroy;
begin
  FDataStrings := nil;
  FRuntimeStrings := nil;
  inherited Destroy;
end;

function TBcExpressionDataSegment.DoOnTag(ATag: string; var AValue: string): Boolean;
begin
  Result := False;
  if Assigned(FOnTag) then
    Result := FOnTag(ATag, AValue);
end;

function TBcExpressionDataSegment.GetCount: Integer;
begin
  Result := Length(FDataStrings);
end;

function TBcExpressionDataSegment.GetStrings(Index: Integer): PBcDataString;
begin
  try
    { we pass pointers to the records. Otherwise the compiler will think
      that we want to *copy* the record data instead of reference. }
    if Index >= 0 then
      Result := @FDataStrings[Index]
    else
      Result := @FRuntimeStrings[Abs(Index) - 1]; // -1 = 0, -2 = 1 ..
  finally
  end;
end;

function TBcExpressionDataSegment.GetTagValue(Index: Integer): string;
var
  Ref: Integer;
  P: PBcDataString;
  S: string;
begin
  Result := '';
  P := Strings[Index];
  Ref := P^.Reference;
  if Ref = -2 then
    Result := P^.Value
  else if Ref = -1 then
  begin
    S := P^.Value;
    if S = 'TRUE' then
      Result := 'TRUE'
    else if S = 'FALSE' then
      Result := 'FALSE';
    if DoOnTag(S, Result) then
    begin
      Ref := Add(Result, -2);
      // why I need to use the Strings[Index]^.Reference?
      // P^.Reference does not work even though they should point to same data?
      // Answer: Add() modifies the dynamic array behind Strings[] thus the P pointer
      // *might* point to invalid data. BIG *might* as dynamic array *might* have been
      // relocated after the memory has been reallocated for it. Note, *might* have been
      // relocated.
      Strings[Index]^.Reference := Ref;
    end else
      Result := '';
  end else
    Result := Strings[P.Reference].Value;
end;

function TBcExpressionDataSegment.IndexOfTag(Value: string; var AResult: Integer): Boolean;
var
  I: Integer;
  R: Integer;
begin
  { tags are strings with reference to the data index ("pointers") that is
    reference is -1 or bigger }
  AResult := -1;
  Result := False;
  for I := 0 to Length(FDataStrings) - 1 do
  begin
    R := FDataStrings[I].Reference;
    if (R > -2) and AnsiSameText(FDataStrings[I].Value, Value) then
    begin
      AResult := I;
      Result := True;
      Break;
    end;
  end;
end;

procedure TBcExpressionDataSegment.AddPredefinedRuntimeData;
begin
  { predefined runtime data }
  AddRuntime('', -2); // pad tmp line index 0
  FResultLine := AddRuntime('', -2); // result line
  FVar1 := AddRuntime('', -2); // tmp
  FVar2 := AddRuntime('', -2); // tmp
//  FLineBreak := AddRuntime(#13#10, -2); // line break
end;

function TBcExpressionDataSegment.GetResultValue: string;
begin
  Result := Strings[FResultLine]^.Value;
end;

function TBcExpressionDataSegment.GetRuntimeCount: Integer;
begin
  Result := Length(FRuntimeStrings);
end;

procedure TBcExpressionDataSegment.ClearResultValue;
begin
  Strings[FResultLine]^.Value := '';
end;

{ TBcExpressionCodeSegment }

function TBcExpressionCodeSegment.Add({var} CodeBlock: TBcCodeBlock): Integer;
begin
  Result := Length(FCodeBlocks);
  SetLength(FCodeBlocks, Result + 1);
  FCodeBlocks[Result] := CodeBlock; // copies the record content
end;

function TBcExpressionCodeSegment.AddClr(Index: Integer): Integer;
var C: TBcCodeBlock;
begin
  FillChar(C, SizeOf(C), 0);
  C.OPCode := OPClr;
  C.Index := Index;
  Result := Add(C);
end;

function TBcExpressionCodeSegment.AddConcat(Source1, Source2,
  Target: Integer): Integer;
var C: TBcCodeBlock;
begin
  FillChar(C, SizeOf(C), 0); // add concat opcode
  C.OPCode := OPConcat;
  C.Target := Target;
  C.Source1 := Source1;
  C.Source2 := Source2;
  Result := Add(C);
end;

function TBcExpressionCodeSegment.AddJmp(JmpOffset: Integer): Integer;
var C: TBcCodeBlock;
begin
  FillChar(C, SizeOf(C), 0);
  C.OPCode := OPJmp;
  C.JmpOffset := JmpOffset;
  Result := Add(C);
end;

procedure TBcExpressionCodeSegment.Clear;
begin
  FCodeBlocks := nil;
  FRunPoint := 0;
  FDataSegment.Clear;
end;

constructor TBcExpressionCodeSegment.Create;
begin
  FDataSegment := TBcExpressionDataSegment.Create;
  FRunPoint := 0; // start from first
end;

destructor TBcExpressionCodeSegment.Destroy;
begin
  FreeAndNil(FDataSegment);
  FCodeBlocks := nil;
  inherited Destroy;
end;

function TBcExpressionCodeSegment.Execute: Integer;
begin
  FRunPoint := 0; // start from first
  DataSegment.ClearResultValue;
  Result := ExecuteNext;
  while Result = 0 do
    Result := ExecuteNext;
  case Result of
    -1: { last code executed } ;
    -2: { concat to tag datastring (read-only) }
  end;
end;

function TBcExpressionCodeSegment.ExecuteNext: Integer;
var B: Boolean;
begin
  Result := 0;
  if (FRunPoint < Length(FCodeBlocks)) then
  begin
    case FCodeBlocks[FRunPoint].OPCode of
      OPNil: { nil, nothing };
      OPHalt: Result := FCodeBlocks[FRunPoint].ExitCode;
      OPExit: Result := FCodeBlocks[FRunPoint].Result;
      OPConcat:
        begin
          with FCodeBlocks[FRunPoint] do
          begin
            if FDataSegment.Strings[Target]^.Reference = -2 then
            begin
              FDataSegment.Strings[Target]^.Value :=
                FDataSegment.GetTagValue(Source1) +
                FDataSegment.GetTagValue(Source2);
            end else
              Result := -2;
          end;
        end;
      OPJmp:
        begin
          Inc(FRunPoint, FCodeBlocks[FRunPoint].JmpOffset);
        end;
      OPCmp:
        begin
          with FCodeBlocks[FRunPoint] do
          begin
            if FCaseSensitive then
              B := AnsiSameStr(FDataSegment.GetTagValue(Exp1), FDataSegment.GetTagValue(Exp2))
            else
              B := AnsiSameText(FDataSegment.GetTagValue(Exp1), FDataSegment.GetTagValue(Exp2));
            if B then
            begin
               Inc(FRunPoint, JumpEqual);
            end else
            begin
              Inc(FRunPoint, JumpNotEqual);
            end;
          end;
        end;
      OPClr: // clears datasegment value (data and runtime)
        with FDataSegment.Strings[FCodeBlocks[FRunPoint].Index]^ do
        begin
          if Reference > -1 then

          Reference := -2;
          Value := '';
        end;
    end;
  end else
    Result := -1;
  Inc(FRunPoint, 1);
end;

function TBcExpressionCodeSegment.GetCodeBlocks(
  Index: Integer): PBcCodeBlock;
begin
  Result := @FCodeBlocks[Index];
end;

function TBcExpressionCodeSegment.GetCount: Integer;
begin
  Result := Length(FCodeBlocks);
end;

function TBcExpressionCompiler.Compile: Boolean;
begin
  Result := Assigned(FCodeSegment);
  if Result then
  begin
    FCodeSegment.Clear;
    ProcessLine(FInput);
  end;
end;

function TBcExpressionCompiler.GetCompilePoint: Integer;
begin
  Result := -1;
  if Assigned(FCodeSegment) then
    Result := Length(FCodeSegment.FCodeBlocks);
end;

function TBcExpressionCodeSegment.GetOutput: string;
begin
  Result := FDataSegment.ResultValue;
end;

function TBcExpressionCompiler.ParseEscapedChars(ALine: string): string;
var
  I, R, L: Integer;
  C: Char;
  Escaped: Boolean;

  procedure AddChar(C: Char);
  begin
    case C of
      #10, #13: { skip } Dec(R, 1);
    else
      Result[R] := C;
    end;
  end;

begin
  if Length(ALine) = 0 then
  begin
    Result := '';
    Exit;
  end;
  L := Length(ALine);
  SetLength(Result, L); { assume that escaped char never resolves to more than 2 chars }
  R := 1;
  I := 1;
  Escaped := False;
  while I <= L do
  begin
    C := ALine[I];
    if Escaped then
    begin
      Escaped := False;
      case UpCase(C) of
        'N': { new line }
          begin
            Result[R] := #13;
            Inc(R, 1);
            Result[R] := #10;
          end;
        else { just escaped char }
          AddChar(C);
      end;
    end else if C = FEscapeChar then
    begin
      Escaped := True;
      Dec(R, 1);
    end else
      AddChar(C);
    Inc(R, 1);
    Inc(I, 1);
  end;
  SetLength(Result, R - 1);
end;

function TBcExpressionCompiler.GetCaseSensitive: Boolean;
begin
  Result := CodeSegment.CaseSensitive;
end;

procedure TBcExpressionCompiler.SetCaseSensitive(const Value: Boolean);
begin
  CodeSegment.CaseSensitive := Value;
end;

end.
