{ Unit: BcPageControlLink
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Links TPageControl pages to TTreeView which can be used to navigate
    the pages. TreeView nodes are automatically created from the TPageControl
    TTabSheets. Designed to make TreeView to a navigation component for a
    PageControl with invisible Tabs without need of update TreeView everytime
    PagaControl changes.

    If Tab is not enabled, it is not put to treeview. Images/ImageIndex are
    set, Tab Caption is used as TreeView node text. TTabSheet Tag can be
    used to create identation to the treeview. Pages are put to treeview
    in the page order. If Tag is 0 then it's placed to same level. If Tag value
    is >0 then the page is put as child node. If Tag value is <0 then
    the Page is put to the Parent of the current node level. You can go back
    more than one level. Identation takes effect for the page being added.

    Please note that the treeview is not previewed on design-time. This is
    by design to save dfm resources which the treeview items would
    take otherwise.

  History:

    2001-10-09: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcPageControlLink;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls;

const
  DefaultFullExpand = False;
  DefaultSelectFirst = True;
  DefaultManualBuild = False;

type
  TPageControlLinkChangeEvent = procedure(Sender: TObject; Page: TTabSheet) of object;

  TBcPageControlLink = class(TComponent)
  private
    { Private declarations }
    FPageControl: TPageControl;
    FTreeView: TCustomTreeView;
    FOldTreeViewOnChange: TTVChangedEvent;
    FOldPageControlOnChange: TNotifyEvent;
    FFullExpand: Boolean;
    FSelectFirst: Boolean;

    BuildingTree: Boolean;
    FOnChange: TPageControlLinkChangeEvent;
    FManualBuild: Boolean;
    FOnAfterTreeViewBuild: TNotifyEvent;
    FOnBeforeTreeViewBuild: TNotifyEvent;

    function GetPageControl: TPageControl;
    function GetTreeView: TCustomTreeView;
    procedure SetPageControl(const Value: TPageControl);
    procedure SetTreeView(const Value: TCustomTreeView);
    procedure SetFullExpand(const Value: Boolean);
  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
    procedure TreeViewOnChange(Sender: TObject; Node: TTreeNode);
    procedure PageControlOnChange(Sender: TObject);
    procedure UnsetEvents;
    procedure SetEvents(const Unset: Boolean);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure BuildTreeView;
  published
    { Published declarations }
    property PageControl: TPageControl read GetPageControl write SetPageControl;
    property TreeView: TCustomTreeView read GetTreeView write SetTreeView;
    property FullExpand: Boolean read FFullExpand write SetFullExpand default DefaultFullExpand;
    property SelectFirst: Boolean read FSelectFirst write FSelectFirst default DefaultSelectFirst;
    property ManualBuild: Boolean read FManualBuild write FManualBuild default DefaultManualBuild;

    property OnChange: TPageControlLinkChangeEvent read FOnChange write FOnChange;
    property OnBeforeTreeView: TNotifyEvent read FOnBeforeTreeViewBuild write FOnBeforeTreeViewBuild;
    property OnAfterTreeViewBuild: TNotifyEvent read FOnAfterTreeViewBuild write FOnAfterTreeViewBuild;
  end;

  TCustomTreeViewAccess = class(TCustomTreeView)
  end;

implementation

{ TBcPageControlLink }

procedure TBcPageControlLink.BuildTreeView;
var
  I, ActiveIndex: Integer;
  ParentNode, NewNode: TTreeNode;
  Page: TTabSheet;

  function GetParentNode(Node: TTreeNode; Levels: Integer): TTreeNode;
  var L: Integer;
  begin
    Result := Node;
    L := Levels;
    while Assigned(Result) and (L > 0) do
    begin
      Result := Result.Parent;
      Dec(L, 1);
    end;
  end;

begin
  if csDesigning in ComponentState then
    Exit;

  try
    BuildingTree := True;
    if Assigned(FOnBeforeTreeViewBuild) then
      FOnBeforeTreeViewBuild(Self);

    if Assigned(FTreeView) and Assigned(FPageControl) then
    begin
      if FSelectFirst then
        ActiveIndex := 0
      else
        ActiveIndex := FPageControl.ActivePageIndex;

      TCustomTreeViewAccess(FTreeView).Items.Clear;

      TCustomTreeViewAccess(FTreeView).Images := FPageControl.Images;
      ParentNode := nil;
      NewNode := nil;
      for I := 0 to FPageControl.PageCount - 1 do
      begin
        Page := FPageControl.Pages[I];
        if not Page.Enabled then
          Continue;

        { use tag to ident }
        if Page.Tag > 0 then
          ParentNode := NewNode
        else if (Page.Tag < 0) then
          ParentNode := GetParentNode(NewNode, Abs(Page.Tag) + 1);

        NewNode := TCustomTreeViewAccess(FTreeView).Items.AddChildObject(ParentNode, Page.Caption, Pointer(Page));
        with NewNode do
        begin
          ImageIndex := Page.ImageIndex;
          SelectedIndex := ImageIndex;
          StateIndex := -1;
        end;
      end;

      if FFullExpand then
        FTreeView.FullExpand;

      BuildingTree := False; // allow onchange to execute
      if (TCustomTreeViewAccess(FTreeView).Items.Count > 0) then
        FTreeView.Selected := TCustomTreeViewAccess(FTreeView).Items[ActiveIndex];
      BuildingTree := True;
    end;
  finally
    if Assigned(FOnAfterTreeViewBuild) then
      FOnAfterTreeViewBuild(Self);
    BuildingTree := False;
  end;
end;

constructor TBcPageControlLink.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFullExpand := DefaultFullExpand;
  FSelectFirst := DefaultSelectFirst;
  FManualBuild := DefaultManualBuild;
  BuildingTree := False;
  FTreeView := nil;
  FPageControl := nil;
  FOldTreeViewOnChange := nil;
end;

destructor TBcPageControlLink.Destroy;
begin
  inherited Destroy;
end;

function TBcPageControlLink.GetPageControl: TPageControl;
begin
  Result := FPageControl;
end;

function TBcPageControlLink.GetTreeView: TCustomTreeView;
begin
  Result := FTreeView;
end;

procedure TBcPageControlLink.Loaded;
begin
  SetEvents(True);
  if not FManualBuild then
    BuildTreeView;
  inherited Loaded;
end;

procedure TBcPageControlLink.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if Operation = opRemove then
  begin
    if AComponent = FPageControl then
    begin
      FPageControl := nil;
      FOldPageControlOnChange := nil;
    end else if AComponent = FTreeView then
    begin
      FTreeView := nil;
      FOldTreeViewOnChange := nil;
    end;
  end;
  inherited Notification(AComponent, Operation);
end;

procedure TBcPageControlLink.PageControlOnChange(Sender: TObject);
begin
  if Assigned(FTreeView) and Assigned(FPageControl) then
  begin
    if (TCustomTreeViewAccess(FTreeView).Items.Count >
      FPageControl.ActivePageIndex) then
      FTreeView.Selected :=
        TCustomTreeViewAccess(FTreeView).Items[FPageControl.ActivePageIndex];
        
    if Assigned(FOnChange) then
      FOnChange(Self, FPageControl.ActivePage);
  end;
  if Assigned(FOldPageControlOnChange) then
    FOldPageControlOnChange(Sender);
end;

procedure TBcPageControlLink.SetEvents(const Unset: Boolean);
begin
  if csDesigning in ComponentState then
    Exit;

  if Unset then
    UnsetEvents;
  if Assigned(FTreeView) then
  begin
    FOldTreeViewOnChange := TCustomTreeViewAccess(FTreeView).OnChange;
    TCustomTreeViewAccess(FTreeView).OnChange := TreeViewOnChange;
  end;
  if Assigned(FPageControl) then
  begin
    FOldPageControlOnChange := FPageControl.OnChange;
    FPageControl.OnChange := PageControlOnChange;
  end;
end;

procedure TBcPageControlLink.SetFullExpand(const Value: Boolean);
begin
  FFullExpand := Value;
end;

procedure TBcPageControlLink.SetPageControl(const Value: TPageControl);
begin
  UnsetEvents;
  FPageControl := Value;
  SetEvents(False);
  if not FManualBuild then
    BuildTreeView;
end;

procedure TBcPageControlLink.SetTreeView(const Value: TCustomTreeView);
begin
  UnsetEvents;
  FTreeView := Value;
  SetEvents(False);
  if not FManualBuild then
    BuildTreeView;
end;

procedure TBcPageControlLink.TreeViewOnChange(Sender: TObject;
  Node: TTreeNode);
begin
  if BuildingTree then
    Exit;

  if Assigned(FTreeView) and Assigned(FPageControl) then
  begin
    if Assigned(Node.Data) then
    begin
      FPageControl.ActivePage := TTabSheet(Node.Data);

      if Assigned(FOnChange) then
        FOnChange(Self, FPageControl.ActivePage);
    end;
  end;

  if Assigned(FOldTreeViewOnChange) then
    FOldTreeViewOnChange(Sender, Node);
end;

procedure TBcPageControlLink.UnsetEvents;
begin
  if csDesigning in ComponentState then
    Exit;

  if Assigned(FTreeView) then
  begin
    TCustomTreeViewAccess(FTreeView).OnChange := FOldTreeViewOnChange;
  end;
  if Assigned(FPageControl) then
  begin
    FPageControl.OnChange := FOldPageControlOnChange;
  end;
end;

end.
